import argparse
import os

from pytest_smell.processing import detect_smells

parser = argparse.ArgumentParser()
parser.add_argument('--verbose', action='store_true',
                    help='Full output which contains smelly tests positions and explanations.')
parser.add_argument('--ci', action='store_true',
                    help='Returns exit code 1 if any smells were found, 0 if the test code is smell free.')
parser.add_argument('--tests_path', type=str,
                    help='Optionally specify path to tests root directory. Default is current directory.')
parser.add_argument('--out_path', type=str,
                    help='Optionally specify path to write the resulting CSV file. Default is current directory.')
args = parser.parse_args()


def pytest_smell():
    test_path = os.getcwd()
    if args.tests_path:
        test_path = args.tests_path

    all_smells = detect_smells(test_path, args.out_path, args.verbose, args.ci)
    if args.ci:
        if not all_smells:
            exit(0)
        else:
            exit(1)



from pytest_smell.constants import CONTROL_STATEMENTS
from pytest_smell.utils import is_assertion, get_assertion_method, starts_try_block, is_sleep, is_annotation, \
    is_print, contains_numeric_literals


def check_assertion_roulette(current_test_lines):
    asserts_with_no_explanation = []
    for line in current_test_lines:
        if is_assertion(line) and len(line.split(',')) == 1:
            asserts_with_no_explanation.append(line)
    return asserts_with_no_explanation


def check_duplicate_assert(current_test_lines):
    seen = set()
    duplicate_assertions = [line for line in current_test_lines if
                            is_assertion(line) and line in seen or seen.add(line)]
    return duplicate_assertions


def check_eager_test(current_test_lines):
    distinct_assertion_methods = set()
    for line in current_test_lines:
        if is_assertion(line):
            assertion_method = get_assertion_method(line, current_test_lines)
            if assertion_method:
                distinct_assertion_methods.add(assertion_method)
    return distinct_assertion_methods


def check_conditional_logic(current_test_lines):
    distinct_conditional_statements = set()
    for line in current_test_lines:
        for statement in CONTROL_STATEMENTS:
            if statement in line:
                distinct_conditional_statements.add(statement)
    return distinct_conditional_statements


def check_exception_handling(current_test_lines):
    try_except_blocks = list()
    for idx, line in enumerate(current_test_lines):
        if starts_try_block(line):
            try_except_blocks.append(line)
            # TODO: Add these back if there is a separation of metric count from display lines
            # try_except_block = get_try_except_block(idx, current_test_lines)
            # try_except_blocks = [*try_except_blocks, *try_except_block]
    return try_except_blocks


def check_sleepy_test(current_test_lines):
    sleep_lines = []
    for line in current_test_lines:
        if is_sleep(line):
            sleep_lines.append(line)
    return sleep_lines


def check_redundant_print(current_test_lines):
    print_lines = []
    for line in current_test_lines:
        if is_print(line):
            print_lines.append(line)
    return print_lines


def check_ignored_test(current_test_lines_with_annotations):
    for line in current_test_lines_with_annotations:
        if is_annotation(line, '@pytest.mark.skip(') or is_annotation(line, '@skip'):
            return [line]
    return []


def check_unknown_test(current_test_lines_with_annotations):
    assert_lines = []
    for line in current_test_lines_with_annotations:
        if is_annotation(line, '@pytest.fixture'):  # means it is a test fixture not a test
            return ['dummy_assert']  # TODO: Improve this
        if is_assertion(line) or is_annotation(line, 'with pytest') or is_annotation(line, 'pytest') or \
                is_annotation(line, 'with warnings'):
            assert_lines.append(line)
    return assert_lines


def check_magic_number(current_test_lines):
    assert_with_magic_numbers_lines = []
    for line in current_test_lines:
        if is_assertion(line) and contains_numeric_literals(line):
            assert_with_magic_numbers_lines.append(line)
    return assert_with_magic_numbers_lines



import csv
import inspect

import pytest_smell.smells as smells
from pytest_smell.config import SMELL_INFO_MAPPING
from .parser import traverse_tests_file


def get_printable_link(file=None, line=None):
    """ Print a link in PyCharm to a line in file.
        Defaults to line where this function was called. """
    if file is None:
        file = inspect.stack()[1].filename
    if line is None:
        line = inspect.stack()[1].lineno
    string = f'File "{file}", line {max(line, 1)}'.replace("\\", "/")
    return string


def output_final_results(count_metrics_mapping):
    print(f'The total number of bad smells detected in the test suite containing '
          f'{count_metrics_mapping["total_test_count"] // len(SMELL_INFO_MAPPING.keys())} tests was:'
          f' {sum([count_metrics_mapping[f"{smell_name}_count"] for smell_name in SMELL_INFO_MAPPING.keys()])}\n')

    for smell_code, smell_info in SMELL_INFO_MAPPING.items():
        print(f'{smell_info["display_name"]} count: {count_metrics_mapping[f"{smell_code}_count"]}')


def write_results_to_file(file_path, all_test_smells):
    with open(file_path, 'w', newline='') as csvfile:
        fieldnames = ['test_suite', 'test_name', 'smell', 'metric_value']
        writer = csv.DictWriter(csvfile, fieldnames=fieldnames)

        writer.writeheader()
        writer.writerows(all_test_smells)


def detect_smells(tests_directory, out_path=None, verbose=True, ci=False):
    count_metrics_names = ['total_test_count',
                           *[f'{smell_name}_count' for smell_name in SMELL_INFO_MAPPING.keys()]]
    count_metrics_mapping = {metric_name: 0 for metric_name in count_metrics_names}
    all_test_smells = []
    for smell_code, smell_info in SMELL_INFO_MAPPING.items():
        is_smelly = False
        for f_path, current_test_name, current_test_lines, line_number in \
                traverse_tests_file(tests_directory, smell_info.get('check_annotations')):
            count_metrics_mapping['total_test_count'] += 1
            smell_handler = getattr(smells, f'check_{smell_code}')
            handler_result = smell_handler(current_test_lines)
            if not smell_info.get('reversed_metric'):
                suffers_of_smell = len(handler_result) > smell_info['metric_threshold']
            else:
                suffers_of_smell = len(handler_result) < smell_info['metric_threshold']
            if suffers_of_smell:
                count_metrics_mapping[f'{smell_code}_count'] += 1
                test_suite_name = f_path.split('/')[-1]
                test_smell_mapping = {'test_suite': test_suite_name, 'test_name': current_test_name,
                                      'smell': smell_info['display_name'], 'metric_value': len(handler_result)}
                all_test_smells.append(test_smell_mapping)
                is_smelly = True
                test_smell_message = f'Test {current_test_name} located at ' \
                                     f'{get_printable_link(f"{tests_directory}/{test_suite_name}", line=line_number)} '\
                                     f'suffers from "{smell_info["display_name"]}". \n' \
                                     f'Explanation: {smell_info["explanation_message"] % len(handler_result)}'
                if verbose and not out_path and not ci:
                    print(test_smell_message)
                if handler_result and verbose and not out_path and not ci:
                    print("\n".join(handler_result))
        if is_smelly and verbose and not out_path and not ci:
            print('\n')

    if not verbose and not out_path and not ci:
        output_final_results(count_metrics_mapping)
    if out_path and not ci:
        write_results_to_file(f'{out_path}/smells.csv', all_test_smells)

    return all_test_smells



import pathlib
from setuptools import setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="pytest-smell",
    version="1.0.1",
    description="Automated bad smell detection tool for Pytest",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/maxpacs98/disertation",
    author="Bodea Alex",
    author_email="alexandrubodeag@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
    ],
    packages=["pytest_smell"],
    entry_points={
        "console_scripts": [
            'pytest-smell=pytest_smell:pytest_smell',
        ]
    },
)
